#include <sys/io.h>
#include <unistd.h>
#include <err.h>

#include "lpt.h"

#define LPT_EOFS	0x0400
#define LPT_STATUS	0x0001
#define LPT_CONTROL	0x0002
#define LPT_ECR		0x0402

#define GET_BIT(d, n)		(((d)>>(n)) & 1)
#define SET_BIT(d, n, b)	(((d)&~(1<<(n)))|(!!(b)<<(n)))

bool lpt::init() {
	if (ioperm(port, 3, 1)) {
		warn("ioperm for 0x%04X", port);
		return false;
	}
	if (ioperm(port + LPT_EOFS, 3, 1)) {
		warn("ioperm for 0x%04X+EOFS", port);
		return false;
	}

	c_control = inb(port + LPT_CONTROL);
	c_ecr = inb(port + LPT_ECR);
	strobe(false);

	return true;
}

unsigned char lpt::ecr(unsigned char ecr) {
	unsigned char tmp = c_ecr;
	outb(c_ecr = ecr, port + LPT_ECR);
	return tmp;
}

unsigned char lpt::status() {
	return inb(port + LPT_STATUS);
}

unsigned char lpt::control(unsigned char ctrl) {
	unsigned char tmp = c_control;
	outb(c_control = ctrl, port + LPT_CONTROL);
	return tmp;
}

unsigned char lpt::data() {
	return inb(port);
}

void lpt::data(unsigned char data) {
	outb(data, port);
}

lpt_mode lpt::mode() const {
	return lpt_mode((ecr() & 0xE0) >> 5);
}

lpt_mode lpt::mode(lpt_mode mode) {
	lpt_mode tmp = this->mode();
	ecr((ecr() & 0x1F) | (mode << 5));
	return tmp;
}

bool lpt::output_enable() const {
	return !GET_BIT(control(), 5);
}

bool lpt::output_enable(bool oe) {
	bool tmp = output_enable();
	control(SET_BIT(control(), 5, !oe));
	return tmp;
}

bool lpt::busy() {
	return GET_BIT(status(), 7);
}

bool lpt::ack() {
	return GET_BIT(status(), 6);
}

bool lpt::paper_out() {
	return GET_BIT(status(), 5);
}

bool lpt::select_in() {
	return GET_BIT(status(), 4);
}

bool lpt::error() {
	return GET_BIT(status(), 3);
}

bool lpt::irq() {
	return !GET_BIT(status(), 2);
}

bool lpt::strobe() {
	return GET_BIT(control(), 0);
}

void lpt::strobe(bool s) {
	control(SET_BIT(control(), 0, s));
}

void lpt::strobe_pulse(int wait) {
	strobe(true);
	if (wait > 0)
		usleep((unsigned int)wait * 1000);
	strobe(false);
}

bool lpt::select_printer() const {
	return GET_BIT(control(), 3);
}

bool lpt::select_printer(bool s) {
	bool tmp = select_printer();
	control(SET_BIT(control(), 3, s));
	return tmp;
}

bool lpt::init_printer() const {
	return GET_BIT(control(), 2);
}

bool lpt::init_printer(bool i) {
	bool tmp = init_printer();
	control(SET_BIT(control(), 2, i));
	return tmp;
}

bool lpt::auto_linefeed() const {
	return GET_BIT(control(), 1);
}

bool lpt::auto_linefeed(bool alf) {
	bool tmp = auto_linefeed();
	control(SET_BIT(control(), 1, alf));
	return tmp;
}
