#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <err.h>

#include <ihx.h>

#include "flashpgm.h"
#include "am29f0x0.h"

static int binary = 0;

static const char optstring[] = "s:r:eEv:S:w:f:h";
static struct option longopts[] = {
	{"size", required_argument, NULL, 's'},
	{"read", required_argument, NULL, 'r'},
	{"write", required_argument, NULL, 'w'},
	{"erase-sector", no_argument, NULL, 'e'},
	{"erase-chip", no_argument, NULL, 'E'},
	{"sector", required_argument, NULL, 'S'},
	{"verify", required_argument, NULL, 'v'},
	{"binary", no_argument, &binary, 1},
	{"ihx", no_argument, &binary, 0},
	{"from", required_argument, NULL, 'f'},
	{"help", no_argument, NULL, 'h'},
	{NULL, 0, NULL, 0}
};

static flashpgm fp;
static memchip *gmc = NULL;

static void help(const char *exe) {
	fprintf(stderr, "Usage: %s [option/s]\n", exe);
	puts("Valid options:");
	puts(" --from, -f          from where to read/write");
	puts(" --sector, -S        from which sector to read/write");
	puts(" --size, -s          size to read/write");
	puts(" --read, -r          read data to specified file");
	puts(" --write, -w         write data from specified file");
	puts(" --erase-sector, -e  erase sector");
	puts(" --erase-chip        erase all data");
	puts(" --verify, -v        verify data with specified file");
	puts(" --ihx               read file as Intel HEX (default)");
	puts(" --binary            read file as raw binary");
	puts(" --help, -h          show this help");
}

static bool init() {
	if (!fp.preinit())
		return false;

	if (!fp.detect(4)) {
		warnx("Programmer not detected!");
		return false;
	}

	fp.init();
	return true;
}

static memchip * detectmem() {
	{
		am29f0x0 *flash = new am29f0x0(&fp);

		if (flash->detect()) {
			printf("Detektiran %s! :)\n", flash->device_name());
			return flash;
		}
	}

	return NULL;
}

static memchip * getmem() {
	if (gmc) return gmc;
	memchip *mc = detectmem();
	if (!mc) errx(1, "Memory chip not detected!");
	return gmc = mc;
}

static unsigned long xstrtoul(const char *s) {
	unsigned long l;
	char *e;
	l = strtoul(s, &e, 0);
	if (e != s + strlen(s))
		errx(1, "Expected number, got \"%s\"!", s);
	return l;
}

int main(int argn, char *argv[]) {
	Buffer buf;
	int opt, lidx;
	memchip *mc;
	unsigned long size = -1U, from_addr = 0;
	bool donesmth = false;

	if (!init())
		return 1;

	while ((opt = getopt_long(argn, argv, optstring, longopts, &lidx)) != -1) {
		switch (opt) {
			case 's':
				size = xstrtoul(optarg);
				break;

			case 'f':
				from_addr = xstrtoul(optarg);
				break;

			case 'r':
				{
				Buffer buf;
				mc = getmem();
				buf.AddElement(mc->read(from_addr, size));
				if (Buffer::Error e = buf.WriteBIN(optarg))
					errx(1, Buffer::GetErrorString(e));
				}
				donesmth = true;
				break;

			case 'v':
			case 'w':
				{
				Buffer buf;
				Buffer::Error e;
				mc = getmem();
				e = !binary ? buf.ReadIHX(optarg) : buf.ReadBIN(optarg, from_addr);
				if (e) errx(1, Buffer::GetErrorString(e));
				if (opt == 'w') {
					if (!mc->write(buf))
						errx(1, "Writing error!");
				} else {
					if (mc->verify(buf))
						exit(1);
				}
				donesmth = true;
				break;
				}

			case 'S':
				mc = getmem();
				from_addr = mc->sector_address(xstrtoul(optarg));
				if (from_addr == -1U)
					errx(1, "Nonexistent sector %s!", optarg);
				break;

			case 'e':
				mc = getmem();
				if (!mc->erase_sector(from_addr))
					errx(1, "Error erasing sector!");
				donesmth = true;
				break;

			case 'E':
				mc = getmem();
				if (!mc->erase_all())
					errx(1, "Error erasing whole memory!");
				donesmth = true;
				break;

			case 0:
				break;

			case 'h':
				help(*argv);
				return 0;

			default:
				help(*argv);
				return 1;
		}
	}

	if (!donesmth) {
		warnx("Nothing done.");
	}

	if (gmc) delete gmc;

	return 0;
}
