#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/time.h>

#include <ihx.h>

#include "memchip.h"

#define BLINK_INTERVAL	344

BuffElem memchip::read(unsigned long from, unsigned long size) {
	if (!space()) return BuffElem();

	if (size > space()) size = space();
	if (size + from > space()) size = space() - from;

	BuffElem be(size);
	unsigned char *p = be.GetBuffer();
	unsigned long lsize = size;

	init_progress(PRGT_READ, "Reading from 0x%lX, 0x%lX bytes...", from, size);
	while (size--) {
		*(p++) = fp->read_at(from++);
		show_progress(100 - size * 100 / lsize);
	}
	done_progress();

	return be;
}

bool memchip::write(const Buffer &buf) {
	unsigned int size = buf.CalcSize(), wd = 0;

	init_progress(PRGT_WRITE, "Writing 0x%X bytes...", size);
	for (Buffer::const_iterator i = buf.begin(); i != buf.end(); ++i) {
		const unsigned char *p = i->GetBuffer();
		for (unsigned int l = 0; l < i->GetSize(); ++l) {
			fp->write_at(i->address + l, p[l]);
			show_progress(++wd * 100 / size);
		}
	}
	done_progress();
	return true;
}

bool memchip::verify(const Buffer &buf) {
	unsigned int size = buf.CalcSize(), vb = 0;

	init_progress(PRGT_READ, "Verifying 0x%X bytes...", size);
	for (Buffer::const_iterator i = buf.begin(); i != buf.end(); ++i) {
		const unsigned char *p = i->GetBuffer();
		for (unsigned int l = 0; l < i->GetSize(); ++l, ++p) {
			unsigned char x = fp->read_at(i->address + l);
			if (x != *p) {
				done_progress(false);
				fprintf(stderr, "Error at 0x%X! Read 0x%02X instead 0x%02X.\n",
						i->address + l, x, *p);
				return false;
			}
			show_progress(++vb * 100 / size);
		}
	}
	done_progress(true);
	return true;
}

unsigned long memchip::sector_address(int id) const {
	return (id == 0) ? space() : -1U;
}

bool memchip::erase_all() {
	unsigned long a = 0, s = space(), ls = s;

	init_progress(PRGT_ERASE, "Erasing...");
	while (s--) {
		fp->write_at(a++, 0x00);
		show_progress(100 - s * 100 / ls);
	}
	done_progress();
	return true;
}

void memchip::show_progress(int percent) {
	if (abs(ltime - msec_time()) >= BLINK_INTERVAL) {
		ltime = msec_time();
		leds_progress(ltype, lleds = !lleds);
	}
	if (lper == percent) return;
	lper = percent;
	printf("\r%d%% done...", percent);
	fflush(stdout);
}

void memchip::leds_progress(prgtype type, bool on) {
	if (!on) {
		fp->led(false, false);
		return;
	}

	switch (type) {
		case PRGT_DONE:
			fp->led(true, false);
			break;
		case PRGT_ERASE:
			fp->led(false, true);
			break;
		case PRGT_READ:
			fp->led(true, false);
			break;
		case PRGT_WRITE:
			fp->led(true, true);
			break;
	}
}

void memchip::init_progress(prgtype type, const char *str, ...) {
	leds_progress(type);
	lper = -1;
	ltime = msec_time();
	lleds = true;
	ltype = type;

	va_list ap;
	va_start(ap, str);
	vprintf(str, ap);
	va_end(ap);
	putchar('\n');
}

void memchip::done_progress(bool ok) {
	if (ok) {
		printf("\rAll done!    \n");
	} else {
		printf("\nError!\n");
	}
	leds_progress(PRGT_DONE);
}

void memchip::info_progress(const char *str, ...) {
	va_list ap;
	putchar('\n');
	va_start(ap, str);
	vprintf(str, ap);
	va_end(ap);
	putchar('\n');
}

int memchip::msec_time() {
	struct timeval tv;
	gettimeofday(&tv, NULL);
	return tv.tv_usec / 1000;
}
